import numpy as np
import netCDF4 as nc
import datetime
import pathlib
import cftime
import pickle

from sscws.sscws import SscWs
from sscws.coordinates import CoordinateComponent, CoordinateSystem, \
    SurfaceGeographicCoordinates

import geospacelab.cs as gsl_cs
from geospacelab import preferences as pfr
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.__dataset_base__ import DatasetSourced

default_dataset_attrs = {
    'data_root_dir': pfr.datahub_data_root_dir / 'SSCWS',
}

default_variable_names = [
                'SC_GEO_LAT', 'SC_GEO_LON', 'SC_GEO_ALT',
                'SC_GEO_X', 'SC_GEO_Y', 'SC_GEO_Z',
                'SC_GSE_X', 'SC_GSE_Y', 'SC_GSE_Z',
                'SC_AACGM_LAT', 'SC_AACGM_LON', 'SC_AACGM_MLT',
                'SC_DATETIME'
            ]


class OrbitPosition_SSCWS(DatasetSourced):

    def __init__(
            self,
            dt_fr: datetime.datetime = None,
            dt_to: datetime.datetime = None,
            sat_id: str = None,
            to_AACGM: bool = True,
            to_netcdf: bool = True,
            time_clip: bool = True,
            allow_download: bool = True,
            force_download: bool = False,
            **kwargs
    ):
        kwargs.update(**default_dataset_attrs)
        super().__init__(**kwargs)

        self.ssc = SscWs()
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.sat_id = sat_id
        self.to_AACGM = to_AACGM
        self.to_netcdf = to_netcdf
        self.time_clip = time_clip
        self.force_download = force_download
        self.allow_download = allow_download
        self.load_data()

    def load_data(self, **kwargs):
        self.check_data_files(load_mode='AUTO', **kwargs)
        self._set_default_variables(
            default_variable_names,
            configured_variables={}
        )

        for file_path in self.data_file_paths:
            fnc = nc.Dataset(file_path)
            variables = {}
            variable_names = [
                'SC_GEO_LAT', 'SC_GEO_LON', 'SC_GEO_ALT',
                'SC_GEO_X', 'SC_GEO_Y', 'SC_GEO_Z',
                'SC_GSE_X', 'SC_GSE_Y', 'SC_GSE_Z',
                'SC_AACGM_LAT', 'SC_AACGM_LON', 'SC_AACGM_MLT',
                'UNIX_TIME'
            ]
            for var_name in variable_names:
                variables[var_name] = np.array(fnc[var_name]).reshape((fnc[var_name].shape[0], 1))

            time_units = fnc['UNIX_TIME'].units

            variables['SC_DATETIME'] = cftime.num2date(variables['UNIX_TIME'].flatten(),
                                                    units=time_units,
                                                    only_use_cftime_datetimes=False,
                                                    only_use_python_datetimes=True)

            variables['SC_DATETIME'] = np.reshape(variables['SC_DATETIME'], (fnc['UNIX_TIME'].shape[0], 1))
            fnc.close()

            for var_name in self._variables.keys():
                value = variables[var_name]
                self._variables[var_name].join(value)

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range(var_datetime_name='SC_DATETIME')

    def search_data_files(self, **kwargs):

        dt_fr = self.dt_fr
        dt_to = self.dt_to

        diff_months = dttool.get_diff_months(dt_fr, dt_to)

        for nm in range(diff_months + 1):

            thismonth = dttool.get_next_n_months(self.dt_fr, nm)

            initial_file_dir = kwargs.pop(
                'initial_file_dir', self.data_root_dir / self.sat_id.upper()
            )

            file_patterns = [
                self.sat_id.upper(),
                thismonth.strftime('%Y%m'),
            ]
            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]
            search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir,
                search_pattern=search_pattern,
                allow_multiple_files=False,
            )
            # Validate file paths

            if (not done and self.allow_download) or self.force_download:
                done = self.download_data(thismonth=thismonth)
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir,
                        search_pattern=search_pattern,
                        allow_multiple_files=True
                    )

        return done

    def download_data(self, thismonth=None):
        done = False
        # diff_months = (self.dt_to.year - self.dt_fr.year) * 12 + (self.dt_to.month - self.dt_fr.month) % 12
        # mylog.simpleinfo.info("Searching the orbit data from NASA/SscWs ...")
        # for nm in range(diff_months + 1):
        #     thismonth = dttool.get_next_n_months(self.dt_fr, nm)
        #     thismonthend = dttool.get_last_day_of_month(thismonth, end=True)
        #     dt_fr_str = thismonth.strftime('%Y-%m-%dT%H:%M:%SZ')
        #     dt_to_str = thismonthend.strftime('%Y-%m-%dT%H:%M:%SZ')
        #
        #     result = self.ssc.get_locations(
        #         [self.sat_id],
        #         [dt_fr_str, dt_to_str], [CoordinateSystem.GEO, CoordinateSystem.GSE]
        #     )
        #
        #     if not list(result['Data']):
        #         return None
        #
        #     data = result['Data'][0]
        #     coords = data['Coordinates'][0]
        #     coords_gse = data['Coordinates'][1]
        #     dts = data['Time']
        #
        #     coords_in = {'x': coords['X'] / 6371.2, 'y': coords['Y'] / 6371.2, 'z': coords['Z'] / 6371.2}
        #     cs_car = gsl_cs.GEOCCartesian(coords=coords_in, ut=dts)
        #     cs_sph = cs_car.to_spherical()
        #     orbits = {
        #         'SC_GEO_LAT': cs_sph['lat'],
        #         'SC_GEO_LON': cs_sph['lon'],
        #         'SC_GEO_ALT': cs_sph['height'],
        #         'SC_GEO_X': coords['X'],
        #         'SC_GEO_Y': coords['Y'],
        #         'SC_GEO_Z': coords['Z'],
        #         'SC_GSE_X': coords_gse['X'],
        #         'SC_GSE_Y': coords_gse['Y'],
        #         'SC_GSE_Z': coords_gse['Z'],
        #         'SC_DATETIME': dts,
        #     }
        #     if self.to_AACGM:
        #         cs_aacgm = cs_sph.to_AACGM(append_mlt=True)
        #         orbits.update(
        #             **{
        #                 'SC_AACGM_LAT': cs_aacgm['lat'],
        #                 'SC_AACGM_LON': cs_aacgm['lon'],
        #                 'SC_AACGM_MLT': cs_aacgm['mlt'],
        #             }
        #         )
        #     if self.to_netcdf:
        #         self.save_to_netcdf(orbits, dt_fr=thismonth, dt_to=thismonthend)
        #     done = True
        mylog.simpleinfo.info("Searching the orbit data from NASA/SscWs ...")
        thismonthend = dttool.get_last_day_of_month(thismonth, end=True)
        dt_fr_str = thismonth.strftime('%Y-%m-%dT%H:%M:%SZ')
        dt_to_str = thismonthend.strftime('%Y-%m-%dT%H:%M:%SZ')

        result = self.ssc.get_locations(
            [self.sat_id],
            [dt_fr_str, dt_to_str], [CoordinateSystem.GEO, CoordinateSystem.GSE]
        )

        if not list(result['Data']):
            return False

        data = result['Data'][0]
        coords = data['Coordinates'][0]
        coords_gse = data['Coordinates'][1]
        dts = data['Time']

        coords_in = {'x': coords['X'] / 6371.2, 'y': coords['Y'] / 6371.2, 'z': coords['Z'] / 6371.2}
        cs_car = gsl_cs.GEOCCartesian(coords=coords_in, ut=dts)
        cs_sph = cs_car.to_spherical()
        orbits = {
            'SC_GEO_LAT': cs_sph['lat'],
            'SC_GEO_LON': cs_sph['lon'],
            'SC_GEO_ALT': cs_sph['height'],
            'SC_GEO_X': coords['X'],
            'SC_GEO_Y': coords['Y'],
            'SC_GEO_Z': coords['Z'],
            'SC_GSE_X': coords_gse['X'],
            'SC_GSE_Y': coords_gse['Y'],
            'SC_GSE_Z': coords_gse['Z'],
            'SC_DATETIME': dts,
        }
        if self.to_AACGM:
            cs_aacgm = cs_sph.to_AACGM(append_mlt=True)
            orbits.update(
                **{
                    'SC_AACGM_LAT': cs_aacgm['lat'],
                    'SC_AACGM_LON': cs_aacgm['lon'],
                    'SC_AACGM_MLT': cs_aacgm['mlt'],
                }
            )
        if self.to_netcdf:
            self.save_to_netcdf(orbits, dt_fr=thismonth, dt_to=thismonthend)
        done = True
        return done

    def save_to_netcdf(self, orbits, dt_fr=None, dt_to=None):
        sat_info = self.get_sat_info(self.sat_id)
        fp = self.data_root_dir / sat_info['Id'].upper() / \
             f"SSCWS_orbits_{sat_info['Id'].upper()}_{dt_fr.strftime('%Y%m')}_{str(sat_info['Resolution']) + 's'}.nc"
        fp.parent.resolve().mkdir(parents=True, exist_ok=True)
        fnc = nc.Dataset(fp, 'w')
        fnc.createDimension('UNIX_TIME', orbits['SC_DATETIME'].shape[0])

        fnc.title = f"{sat_info['Name']} orbits from {dt_fr.strftime('%Y%m%dT%H%M%S')} to {dt_to.strftime('%Y%m%dT%H%M%S')}"
        time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
        time.units = 'seconds since 1970-01-01 00:00:00.0'
        geo_lat = fnc.createVariable('SC_GEO_LAT', np.float32, ('UNIX_TIME',))
        geo_lon = fnc.createVariable('SC_GEO_LON', np.float32, ('UNIX_TIME',))
        geo_alt = fnc.createVariable('SC_GEO_ALT', np.float32, ('UNIX_TIME',))
        geo_x = fnc.createVariable('SC_GEO_X', np.float32, ('UNIX_TIME',))
        geo_y = fnc.createVariable('SC_GEO_Y', np.float32, ('UNIX_TIME',))
        geo_z = fnc.createVariable('SC_GEO_Z', np.float32, ('UNIX_TIME',))
        gse_x = fnc.createVariable('SC_GSE_X', np.float32, ('UNIX_TIME',))
        gse_y = fnc.createVariable('SC_GSE_Y', np.float32, ('UNIX_TIME',))
        gse_z = fnc.createVariable('SC_GSE_Z', np.float32, ('UNIX_TIME',))

        time_array = np.array(
            cftime.date2num(orbits['SC_DATETIME'].flatten(), units='seconds since 1970-01-01 00:00:00.0'))
        time[::] = time_array[::]
        geo_lat[::] = orbits['SC_GEO_LAT'][::]
        geo_lon[::] = orbits['SC_GEO_LON'][::]
        geo_alt[::] = orbits['SC_GEO_ALT'][::]
        geo_x[::] = orbits['SC_GEO_X'][::]
        geo_y[::] = orbits['SC_GEO_Y'][::]
        geo_z[::] = orbits['SC_GEO_Z'][::]
        gse_x[::] = orbits['SC_GSE_X'][::]
        gse_y[::] = orbits['SC_GSE_Y'][::]
        gse_z[::] = orbits['SC_GSE_Z'][::]

        if self.to_AACGM:
            aa_lat = fnc.createVariable('SC_AACGM_LAT', np.float32, ('UNIX_TIME',))
            aa_lon = fnc.createVariable('SC_AACGM_LON', np.float32, ('UNIX_TIME',))
            aa_mlt = fnc.createVariable('SC_AACGM_MLT', np.float32, ('UNIX_TIME',))
            aa_lat[::] = orbits['SC_AACGM_LAT'][::]
            aa_lon[::] = orbits['SC_AACGM_LON'][::]
            aa_mlt[::] = orbits['SC_AACGM_MLT'][::]

        mylog.simpleinfo.info(
            f"The {sat_info['Name']} orbit data from" +
            f" {dt_fr.strftime('%Y%m%dT%H%M%S')} to {dt_to.strftime('%Y%m%dT%H%M%S')}" +
            f" has been saved to {fp}."
        )
        fnc.close()

    @staticmethod
    def list_satellites(reload=False, logging=True):
        file_path = pfr.datahub_data_root_dir / 'SSCWS' / 'SSCWS_info_satellites.pkl'
        if not file_path.is_file():
            file_path.parent.resolve().mkdir(exist_ok=True)
            reload = True

        if reload:
            ssc = SscWs()
            res = ssc.get_observatories()
            sat_list = res['Observatory']
            with open(file_path, 'wb') as fobj:
                pickle.dump(sat_list, fobj, pickle.HIGHEST_PROTOCOL)

        with open(file_path, 'rb') as fobj:
            sat_list = pickle.load(fobj)

        if logging:
            mylog.simpleinfo.info(
                '{:<20s}{:<30s}{:<20s}{:<25s}{:<25s}{:80s}'.format(
                    'ID', 'NAME', 'RESOLUTION (s)', 'FROM', 'TO', 'RESOURCE'
                )
            )
            for sat_info in sat_list:
                id = sat_info['Id']
                name = sat_info['Name']
                res = sat_info['Resolution']
                dt_fr = sat_info['StartTime']
                dt_to = sat_info['EndTime']
                resource = sat_info['ResourceId']

                mylog.simpleinfo.info(
                    '{:<20s}{:30s}{:<20d}{:<25s}{:<25s}{:50s}'.format(
                        id, name, res, dt_fr.strftime('%Y-%m-%d'), dt_to.strftime('%Y-%m-%d'), str(resource)
                    )
                )
        return sat_list

    @staticmethod
    def get_sat_info(sat_id):
        sat_list = OrbitPosition_SSCWS.list_satellites(logging=False)
        sat_ids = [sat_info['Id'] for sat_info in sat_list]
        try:
            ind = sat_ids.index(sat_id)
        except ValueError:
            raise KeyError('The satellite ID does not exist!')

        return sat_list[ind]

    @staticmethod
    def list_stations(reload=False, logging=True):
        file_path = pfr.datahub_data_root_dir / 'SSCWS' / 'SSCWS_info_ground_stations.pkl'
        if not file_path.is_file():
            file_path.parent.resolve().mkdir(exist_ok=True)
            reload = True

        if reload:
            ssc = SscWs()
            res = ssc.get_ground_stations()
            station_list = res['GroundStation']
            with open(file_path, 'wb') as fobj:
                pickle.dump(station_list, fobj, pickle.HIGHEST_PROTOCOL)

        with open(file_path, 'rb') as fobj:
            station_list = pickle.load(fobj)

        if logging:
            mylog.simpleinfo.info(
                '{:^20s}{:30s}{:30s}{:30s}'.format(
                    'ID', 'NAME', 'LATITUDE', 'LONGITUDE'
                )
            )
            for station_info in station_list:
                id = station_info['Id']
                name = station_info['Name']
                lat = station_info['Location']['Latitude']
                lon = station_info['Location']['Longitude']

                mylog.simpleinfo.info(
                    '{:^20s}{:30s}{:<30.2f}{:<30.2f}'.format(
                        id, name, lat, lon 
                    )
                )
        return station_list

    @staticmethod
    def get_station_info(station_id):
        station_list = OrbitPosition_SSCWS.list_satellites(logging=False)
        station_ids = [sat_info['Id'] for sat_info in station_list]
        try:
            ind = station_ids.index(station_id)
        except ValueError:
            raise KeyError('The satellite ID does not exist!')

        return station_list[ind]

    @staticmethod
    def calc_orbit_az_el(lat, lon, height):
        cs_geo = gsl_cs.GEOCSpherical(coords={'lat': lat, 'lon': lon, 'height':height})
        
        cs_geo = cs_geo.to_cartesian()

    def add_GEO_LST(self):
        lons = self['SC_GEO_LON'].flatten()
        uts = self['SC_DATETIME'].flatten()
        lsts = [ut + datetime.timedelta(hours=lon/15.) for ut, lon in zip(uts, lons)]
        lsts = [lst.hour + lst.minute/60. + lst.second/3600. for lst in lsts]
        var = self.add_variable(var_name='SC_GEO_LST')
        var.value = np.array(lsts)[:, np.newaxis]
        var.label = 'LST'
        var.unit = 'h'
        var.depends = self['SC_GEO_LON'].depends
        return var

if __name__ == "__main__":
    dt_fr = datetime.datetime(2016, 1, 1,)
    dt_to = datetime.datetime(2016, 3, 3)
    OrbitPosition_SSCWS.list_satellites()
    orbit_info = OrbitPosition_SSCWS(dt_fr, dt_to, sat_id='swarma')


"""

The module datahub is the data manager in GeospaceLab. The datahub has three core class-based components:

    - :class:`DataHub <geospacelab.datahub.DataHub>` manages a set of datasets docked or added to the datahub.
    - :class:`Dataset <geospacelab.datahub.DatasetModel>` manages a set of variables loaded from a data source.
    - :class:`Variable <geospacelab.datahub.VariableModel>` records the value, error, and various attributes \
    (e.g., name, label, unit, depends, ndim, ...) of a variable.

"""

# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

# __all__ = ['DataHub', 'DatasetModel', 'VariableModel',]

import datetime
import pathlib
import numpy as np

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pybasic as pybasic
import geospacelab.toolbox.utilities.pyclass as pyclass
from geospacelab import preferences as pref


class DatasetBase(object):
    """
    A dataset is a dictionary-like object used for downloading and loading data from a data source. The items in
    the dataset are the variables loaded from the data files. The parameters listed below are the general attributes used
    for the dataset class and its inheritances.

    :ivar str name: The name of the dataset.
    :ivar str kind: The type of the dataset. 'sourced': the data source has been added in the package,
      'temporary': a dataset added temporarily, 'user-defined': a data source defined by the user.
    :ivar datetime.datetime or None dt_fr: the starting time of the data records.
    :ivar datetime.datetime or None dt_fr: the starting time of the data records.
    :ivar str, {"on", "off"} visual: If "on", append the Visual object to the Variable object.
    :ivar list label_fields: A list of strings, indicating the fields used for generating the dataset label.
    """

    from geospacelab.datahub.__variable_base__ import VariableBase as VariableModel
    __variable_model__ = VariableModel

    def __init__(self,
                 dt_fr: datetime.datetime = None,
                 dt_to: datetime.datetime = None,
                 name: str = '',
                 kind: str = '',
                 visual: str = 'off',
                 label_fields: list = ('name', 'kind'), **kwargs):
        """
        Initial inputs to create the object.

            :param name: The name of the dataset.
            :type name: str or None
            :param kind: The type of the dataset. 'sourced': the data source has been added in the package,
              'temporary': a dataset added temporarily, 'user-defined': a data source defined by the user.
            :type kind: {'sourced', 'temporary', 'user-defined'}
            :param dt_fr: the starting time of the data records.
            :type dt_fr: datetime.datetime, default: None
            :param dt_to: the stopping time of the the data records.
            :type dt_to: datetime.datetime, default: None
            :param visual: If "on", append the attribute visual to the vairables.
            :type visual: {'on', 'off'}
            :param label_fields: A list of strings, indicating the fields used for generating the dataset label.
            :type label_fields: list, {['name', 'kind']}
        """

        self._variables = {}
        self.name = name
        self.kind = kind
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.visual = visual

        self.label_fields = label_fields

    def add_variable(self, var_name: str, configured_variables=None, variable_class=None, **kwargs) -> __variable_model__:
        """
        Add a variable to the dataset.

        :param var_name:
        :param configured_variables:
        :param variable_class:
        :param kwargs:
        :return:
        """
        if variable_class is None:
            variable_class = self.__variable_model__
        if configured_variables is None:
            configured_variables = {}
        if var_name in configured_variables.keys():
            configured_variable = configured_variables[var_name]
            if type(configured_variable) is dict:
                self[var_name] = variable_class(**configured_variable)
            elif issubclass(configured_variable.__class__, self.__variable_model__):
                self[var_name] = configured_variable.clone()
            # self[var_name].dataset = self
            self[var_name].visual = self.visual
        else:
            self[var_name] = variable_class(dataset=self, name=var_name, visual=self.visual, **kwargs)
        return self[var_name]
    
    def clone_variables(self, ds, variable_names=None):
        if not isinstance(ds, DatasetBase):
            raise TypeError
        if variable_names is None:
            variable_names = ds.keys()
        for var_name in variable_names:
            self[var_name] = ds[var_name].clone()
            self[var_name].dataset = self
        return None

    def label(self, fields=None, separator=' | ', lowercase=True, num_to_str=True) -> str:
        """
        Return a label of the data set.
        :param fields: The attribute names for the label.
        :param separator: A separator between two attributes.
        :param lowercase: Show lowercase letters only.
        :return: label
        """
        if fields is None:
            fields = self.label_fields
        sublabels = []
        for attr_name in fields:
            if not str(attr_name):
                sublabels.append('*')
            else:
                attr = getattr(self, attr_name)
                if pybasic.isnumeric(attr) and num_to_str:
                    attr = str(attr)
                sublabels.append(attr)
        label = pybasic.str_join(*sublabels, separator=separator, lowercase=lowercase)
        return label

    def config(self, logging: bool = True, **kwargs) -> None:
        """
        Configure the attributes of the dataset.

        :param logging: Show logging if True.
        :param kwargs:
        :return:
        """
        pyclass.set_object_attributes(self, append=False, logging=logging, **kwargs)

    def add_attr(self, logging=True, **kwargs) -> None:
        pyclass.set_object_attributes(self, append=True, logging=logging, **kwargs)

    def attrs_to_dict(self, attr_names=None) -> dict:
        if attr_names is None:
            attr_names = []
        items = {}
        for attr_name in attr_names:
            items[attr_name] = getattr(self, attr_name)
        return items

    def list_all_variables(self) -> None:
        # print header
        label = self.label()
        mylog.simpleinfo.info("Dataset: {}".format(label))
        mylog.simpleinfo.info("Printing all of the variables ...")
        mylog.simpleinfo.info('{:^20s}{:^30s}'.format('No.', 'Variable name'))
        for ind, var_name in enumerate(self._variables.keys()):
            mylog.simpleinfo.info('{:^20d}{:30s}'.format(ind+1, var_name))
        mylog.simpleinfo.info('')

    def keys(self) -> list:
        return list(self._variables.keys())

    def items(self) -> dict:
        return dict(self._variables.items())

    def __setitem__(self, key, value):
        if value is not None:
            if not issubclass(value.__class__, self.__variable_model__):
                raise TypeError
        self._variables[key] = value
        self._variables[key].dataset = self

    def __getitem__(self, key) -> __variable_model__:
        # if key not in self._variables.keys():
        #    raise KeyError
        return self._variables[key]

    def __delitem__(self, key):
        del self._variables[key]
        pass

    def __repr__(self):
        rep = f"GeospaceLab Dataset object <{self.label()}, " + \
              f"starting time: {str(self.dt_fr)}, stopping time: {str(self.dt_to)}>"
        return rep.replace('geospacelab.datahub.sources.', '')

    # def add_variable(self, variable, name=None):
    #     if issubclass(variable, VariableModel):
    #         pass
    #     else:
    #         variable = VariableModel(value=variable, name=name, visual=self.visual)
    #     self[variable.name] = variable
    def exist(self, var_name) -> bool:
        if var_name in self._variables.keys():
            return True
        else:
            return False

    def remove_variable(self, name):
        del self[name]

    def get_variable_names(self) -> list:
        return list(self._variables.keys())

    def register_method(self, func):
        from types import MethodType
        setattr(self, func.__name__, MethodType(func, self))


class DatasetUser(DatasetBase):
    def __init__(self,
                 dt_fr: datetime.datetime = None,
                 dt_to: datetime.datetime = None,
                 name: str = '',
                 visual: str = 'off',
                 label_fields: list = ('name', 'kind'), **kwargs):

        super().__init__(
            dt_fr=dt_fr, dt_to=dt_to, name=name, kind='user-defined', visual=visual, label_fields=label_fields, **kwargs
        )


class DatasetSourced(DatasetBase):

    def __init__(self,
                 dt_fr: datetime.datetime = None,
                 dt_to: datetime.datetime = None,
                 name: str = '',
                 visual: str = 'off',
                 label_fields: list = ('name', 'kind'), **kwargs):

        kind = kwargs.pop('kind', 'sourced')
        super().__init__(
            dt_fr=dt_fr, dt_to=dt_to, name=name, kind=kind, visual=visual, label_fields=label_fields, **kwargs
        )

        self.loader = kwargs.pop('loader', None)
        self.downloader = kwargs.pop('downloader', None)
        self.load_mode = kwargs.pop('load_mode', 'AUTO')  # ['AUTO'], 'dialog', 'assigned'
        self.data_root_dir = kwargs.pop('data_root_dir', pref.datahub_data_root_dir)
        self.data_file_paths = kwargs.pop('data_file_paths', [])
        self.data_file_num = kwargs.pop('data_file_num', 0)
        self.data_file_ext = kwargs.pop('data_file_ext', '*')
        self.data_search_recursive = kwargs.pop('data_search_recursive', False)
        self.time_clip = kwargs.pop('time_clip', True)

    def search_data_files(
            self,
            initial_file_dir=None, search_pattern='*', recursive=None,
            direct_append=True, allow_multiple_files=False, include_extension=True,
            **kwargs) -> list:
        """
        Search the data files by the input pattern in the file name. The search method is based on pathlib.glob.
        For a dataset inheritance, a wrapper can be added for a custom setting.

        :param initial_file_dir: The initial file directory for searching.
        :type initial_file_dir: str or pathlib.Path, default: DatasetModel.data_root_dir.
        :param search_pattern: Unix style pathname pattern, see also pathlib.glob.
        :type search_pattern: str.
        :param recursive: Search recursively if True.
        :type recursive: bool, default: DatasetModel.data_search_recursive.
        :param allow_multiple_files: Allow multiple files as a result.
        :type allow_mulitple_files: bool, default: False.
        :param direct_append: Append directly the searched results to Dataset.data_file_paths.
            If `False`, the file path list is returned only.
        :type direct_append: bool, default: `True`

        :return: a list of the file paths.
        """

        file_paths = []
        if initial_file_dir is None:
            initial_file_dir = self.data_root_dir
        if recursive is None:
            recursive = self.data_search_recursive
        if include_extension:   
            if str(self.data_file_ext):
                search_pattern = search_pattern + '.' + self.data_file_ext
        if recursive:
            search_pattern = '**/' + search_pattern
        paths = list(initial_file_dir.glob(search_pattern))

        import natsort
        paths = natsort.natsorted(paths, reverse=False)
        if len(paths) == 1:
            file_paths.extend(paths)
        elif len(paths) > 1:
            if allow_multiple_files:
                file_paths.extend(paths)
            else:
                mylog.StreamLogger.error("Multiple files found! Restrict the search condition.")
                print(paths)
                raise FileExistsError
        else:
            mylog.simpleinfo.info('Cannot find the requested data file in {}'.format(initial_file_dir))
        if direct_append:
            self.data_file_paths.extend(file_paths)
        return file_paths

    def open_dialog(self, initial_file_dir: str = None, data_file_num: int = None, **kwargs):
        """
        Open a dialog to select the data files.
        """
        import tkinter as tk
        from tkinter import filedialog
        from tkinter import simpledialog
        def tk_open_file():

            root = tk.Tk()
            root.withdraw()

            dialog_title = "Select a data file ..."

            if str(self.data_file_ext):
                p1 = '*.' + self.data_file_ext
            file_types = (('data files', p1), ('all files', '*.*'))
            file_name = filedialog.askopenfilename(
                title=dialog_title,
                initialdir=initial_file_dir,
                filetypes=file_types,
            )

            return file_name

        root = tk.Tk()
        root.withdraw()
        self.data_file_num = kwargs.pop('data_file_num', self.data_file_num)
        if self.data_file_num == 0:
            self.data_file_num = simpledialog.askinteger('Input dialog', 'Input the number of files:', initialvalue=1)

        for nf in range(self.data_file_num):
            title = f"Select File {nf+1}: "
            file_types = (('eiscat files', '*.' + self.data_file_ext), ('all files', '*.*'))
            file_name = filedialog.askopenfilename(
                title=title,
                initialdir=initial_file_dir
            )
            self.data_file_paths.append(pathlib.Path(file_name))

    def check_data_files(self, load_mode: str = None, **kwargs):
        """
        Check the existing of the data files before loading the data, depending on the loading mode (``load_mode``).
        This methods still needs to be improved as different datasets may have different variables as epochs. Two kinds
        of things can be done: 1. write a wrapper in the new dataset inheritance. 2. Add a script to recognize the
        epoch variables.

        """
        if load_mode is not None:
            self.load_mode = load_mode
        if list(self.data_file_paths):
            self.load_mode = 'assigned'
        if self.load_mode == 'AUTO':
            self.search_data_files(**kwargs)
        elif self.load_mode == 'dialog':
            self.open_dialog(**kwargs)
        elif self.load_mode == 'assigned':
            self.data_file_paths = kwargs.pop('data_file_paths', self.data_file_paths)
            self.data_file_paths = [pathlib.Path(f) for f in self.data_file_paths]
        else:
            raise NotImplementedError
        self.data_file_num = len(self.data_file_paths)

    def time_filter_by_range(self, var_datetime=None, var_datetime_name=None):
        """
        Clip the times.
        :param var_datetime:
        :param var_datetime_name:
        :return:
        """
        if var_datetime is None and var_datetime_name is None:
            var_datetime = self['DATETIME']
        if var_datetime_name is not None:
            var_datetime = self[var_datetime_name]
        if var_datetime.value is None:
            return
        inds = np.where((var_datetime.value.flatten() >= self.dt_fr) & (var_datetime.value.flatten() <= self.dt_to))[0]
        self.time_filter_by_inds(inds, var_datetime=var_datetime)

    def time_filter_by_inds(self, inds, var_datetime=None):
        if inds is None:
            return
        if not list(inds):
            return
        if var_datetime is None:
            var_datetime = self['DATETIME']

        shape_0 = var_datetime.value.shape[0]
        for var in self._variables.values():
            if var.value is None:
                continue
            if var.value.shape[0] == shape_0 and len(var.value.shape) > 1:
                var.value = var.value[inds, ::]

    def get_time_ind(self, ut, time_res=None, var_datetime=None, var_datetime_name=None, edge_cutoff=True):
        if var_datetime is None and var_datetime_name is None:
            var_datetime = self['DATETIME']
        if var_datetime_name is not None:
            var_datetime = self[var_datetime_name]
        if var_datetime.value is None:
            return
        dts = var_datetime.value.flatten()

        if time_res is not None:
            edge_cutoff = False

        ind = []
        if edge_cutoff:
            if ut > dts[-1] or ut < dts[0]:
                mylog.StreamLogger.warning('The input time is out of the range! Set "edge_cutoff=False" if needed!')
                return ind
        delta_sectime = [delta_t.total_seconds() for delta_t in (dts - ut)]
        
        ind = np.where(np.abs(delta_sectime) == np.min(np.abs(delta_sectime)))[0][0]

        if np.abs((dts[ind] - ut).total_seconds()) > time_res:
            mylog.StreamLogger.warning('The input time does not match any time in the list! Check the time resolution ("time_res") in seconds!')
            return []
        return ind

    def _set_default_variables(self, default_variable_names, configured_variables=None):
        if configured_variables is None:
            configured_variables = {}
        for var_name in default_variable_names:
            self.add_variable(var_name, configured_variables=configured_variables)

    @property
    def data_root_dir(self):
        return self._data_root_dir

    @data_root_dir.setter
    def data_root_dir(self, path):
        self._data_root_dir = pathlib.Path(path)

class LoaderBase(object):
    """
    Loader Base for loading data from the assigned files
    """
    def __init__(self):
        self.variables = {}
        self.metadata = {}

    def load_data(self):
        raise NotImplementedError


class DownloaderBase(object):
    """
    Downloader Base for downloading data files from different data sources

    Parameters
    ===========
    :param dt_fr: starting time.
    :type dt_fr: datetime.datetime
    :param dt_to: stopping time.
    :type dt_to: datetime.datetime
    :param data_file_root_dir: the root directory storing the data.
    :type data_file_root_dir: str or pathlib.Path object
    :param done: if the files are downloaded or not [False].
    :type done: bool
    :param force: Force the downloader to download the data files, [True].
    :type force: bool
    :param direct_download: Download the data directly without calling the function download [True].
    :type direct_download: bool
    """

    def __init__(self, dt_fr, dt_to, data_file_root_dir=None, force=True, direct_download=True, **kwargs):

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.data_file_root_dir = data_file_root_dir

        self.files = []

        self.force = force
        self.direct_download = direct_download
        self.done = False

        if self.direct_download:
            self.done = self.download(**kwargs)

    def download(self, **kwargs):
        """
        Download the data from a http or ftp server. Implementation must be added according to different data sources
        :param kwargs: keywords for downloading the data files.
        """
        raise NotImplemented



import datetime
import numpy as np


def solar_declination(dt, degrees=False):
    factor = np.pi / 180.
    sd_SS = - 23.4333333 * factor   # Declination angle at southern summer solstice.

    dt_0 = datetime.datetime(dt.year, dt.month, dt.day, 0, )
    N = (dt - dt_0).total_seconda() / 86400.
    sin_EL = np.cos(
        2 * np.pi / 365.24 * (N + 10) + 2 * 0.0167 * np.sin(2 * np.pi/365.24 * (N - 2))
    )

    sd = np.arcsin(
        np.sin(sd_SS) * sin_EL
    )

    if degrees:
        sd = sd / factor
    return sd


def subsolar_point(dt, degrees=False):
    phi = solar_declination(dt, degrees)

    lamda = - 15 * ((dt.hour + dt.minute / 60 + dt.second / 3600) - 12. + equation_of_time(dt))

    
def equation_of_time(dt):
    return




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as dtparse
import numpy as np
import re
import requests
import bs4
import os
import pathlib

from geospacelab import preferences as pfr
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


def test():
    dt_fr = datetime.datetime(2021, 3, 10)
    dt_to = datetime.datetime(2021, 3, 10)
    download_obj = Downloader(dt_fr, dt_to)


data_type_dict = {
    'TEC-MAP': 'TEC binned',
    'TEC-LOS': 'Line of sight',
    'TEC-Sites': 'List of sites',
}


class Downloader(object):
    """Download the GNSS TEC data
    """

    def __init__(self, dt_fr, dt_to, data_file_root_dir=None, file_type='TEC-MAP', force_download=False):
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.force_download = force_download
        if data_file_root_dir is None:
            self.data_file_root_dir: pathlib.Path = pfr.datahub_data_root_dir / 'ISEE' / 'GNSS' / 'TEC'
        else:
            self.data_file_root_dir = data_file_root_dir
        self.done = False

        self.root_url = "https://stdb2.isee.nagoya-u.ac.jp/GPS/shinbori/AGRID2/nc/"
        self.download()

    def download(self):
        ndays = dttool.get_diff_days(self.dt_fr, self.dt_to)
        dt_0 = dttool.get_start_of_the_day(self.dt_fr)
        for nd in range(ndays + 1):
            thisday = dt_0 + datetime.timedelta(days=nd)
            doy = dttool.get_doy(thisday)
            for hh in range(24):
                this_time = thisday + datetime.timedelta(hours=hh)
                url = self.root_url + \
                    "{:04d}/{:03d}/".format(thisday.year, doy) + \
                    this_time.strftime("%Y%m%d%H") + '_atec.nc'

                file_dir = self.data_file_root_dir / str(thisday.year) / thisday.strftime("%Y%m%d")
                file_dir.mkdir(parents=True, exist_ok=True)
                file_name = this_time.strftime("%Y%m%d%H") + '_atec.nc'
                file_path = file_dir / file_name
                if file_path.is_file():
                    mylog.simpleinfo.info("The file {} has been downloaded.".format(file_path.name))
                    self.done = True
                    continue
                mylog.simpleinfo.info("Downloading  {} from the ISEE GNSS database ...".format(file_name))
                r = requests.get(url)
                with open(file_dir / file_name, 'wb') as f:
                    f.write(r.content)
                self.done = True


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import string
import pathlib
import matplotlib.pyplot as plt
import numpy as np
import datetime
import matplotlib as mpl
import matplotlib.dates as mdates

from geospacelab.datahub import DataHub

import geospacelab.visualization.mpl.panels as mpl_panel
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pydatetime as dttool
from geospacelab.visualization.mpl.__base__ import DashboardBase
import geospacelab.visualization.mpl.panels as panels


class Dashboard(DataHub, DashboardBase):
    _default_layout_config = {
        'left': 0.15,
        'right': 0.78,
        'bottom': 0.15,
        'top': 0.88,
        'hspace': 0.1,
        'wspace': 0.1
    }

    def __init__(self, figure=None, figure_config=None, **kwargs):
        super(Dashboard, self).__init__(figure=figure, figure_config=figure_config, **kwargs)


class TSDashboard(Dashboard):
    _default_figure_config = {
        'figsize': (10, 8),
    }

    def __init__(
            self,
            figure=None, figure_config=None,
            time_gap=True,
            timeline_major='UT',
            timeline_extra_labels=None,
            **kwargs
    ):
        self.panel_layouts = []
        self.plot_styles = None
        self._panels_configs = {}

        figure_config = self._default_figure_config if figure_config is None else figure_config
        self.time_gap = time_gap
        self.timeline_major = timeline_major
        self.timeline_extra_labels = timeline_extra_labels
        kwargs.update(visual='on')

        super(Dashboard, self).__init__(figure=figure, figure_config=figure_config, **kwargs)

        self.panels: dict[int: panels.TSPanel] = {}
        self._xlim = [self.dt_fr, self.dt_to]

    def set_layout(self, panel_layouts=None, panels_classes=None, plot_styles=None, row_height_scales=1,
                   left=None, right=None, bottom=None, top=None, hspace=None, add_panels=True, **kwargs):

        if left is None:
            left = self._default_layout_config['left']
        if right is None:
            right = self._default_layout_config['right']
        if bottom is None:
            bottom = self._default_layout_config['bottom']
        if top is None:
            top = self._default_layout_config['top']
        if hspace is None:
            hspace = self._default_layout_config['hspace']

        num_rows = len(panel_layouts)
        self.panel_layouts = panel_layouts
        if type(plot_styles) is not list:
            self.plot_styles = [None] * num_rows
        elif len(plot_styles) != num_rows:
            raise ValueError

        if panels_classes is None:
            panels_classes = [panels.TSPanel] * num_rows
        elif issubclass(panels_classes, panels.TSPanel):
            panels_classes = [panels_classes] * num_rows
        elif len(panels_classes) != num_rows:
            raise ValueError
        else:
            raise AttributeError

        if type(row_height_scales) is not list:
            if type(row_height_scales) == int:
                row_height_scales = [row_height_scales] * num_rows
            else:
                raise TypeError
        elif len(row_height_scales) != num_rows:
            raise ValueError

        hspace = hspace * row_height_scales[0]
        gs_num_rows = sum(row_height_scales)
        gs_num_cols = 1
        super().set_layout(num_rows=gs_num_rows, num_cols=gs_num_cols, left=left, right=right, bottom=bottom, top=top,
                           hspace=hspace, **kwargs)
        rec = 0
        for ind, height in enumerate(row_height_scales):
            self._panels_configs[ind] = {
                'row_ind': [rec, rec+height],
                'col_ind': [0, 1],
                'panel_class': panels_classes[ind],
            }
            rec = rec + height
        if add_panels:
            self.add_panels()

    def add_panels(self):
        bottom_panel = False
        for ind, panel_config in self._panels_configs.items():
            # share x
            if ind > 0:
                panel_config.update(sharex=self.panels[0]())
            if ind == len(self._panels_configs.keys())-1:
                bottom_panel = True
                panel_config.update(timeline_extra_labels=self.timeline_extra_labels)
            else:
                bottom_panel = False
            panel_config.update(bottom_panel=bottom_panel)
            self.add_panel(**panel_config)

    def draw(self, dt_fr=None, dt_to=None, auto_grid=True, time_res=None):
        if dt_fr is not None:
            self._xlim[0] = dt_fr
        if dt_to is not None:
            self._xlim[1] = dt_to

        if not isinstance(time_res,list):
            time_res = [time_res] * len(self.panels.keys())

        npanel = 0
        for ind, panel in self.panels.items():
            panel._xlim = self._xlim
            panel.time_res = time_res[ind]
            plot_layout = self.panel_layouts[npanel]
            plot_layout = self._validate_plot_layout(plot_layout)
            panel.draw(plot_layout)
            npanel = npanel + 1

        if auto_grid:
            self.add_grid(panel_id=0, visible=True, which='major', axis='both', lw=0.5, color='grey', alpha=0.3)
            self.add_grid(panel_id=0, visible=True, which='minor', axis='both', lw=0.3, color='grey', alpha=0.1)

    def add_grid(self, panel_id=0, visible=None, which='major', axis='both', **kwargs):
        if panel_id == 0:
            panels = self.panels.values()
        elif type(panel_id) is int:
            panels = [self.panels[int]]
        elif type(panel_id) is list:
            panels = [self.panels[i] for i in panel_id]
        else:
            raise NotImplementedError

        for p in panels:
            p.add_grid(ax=p(), visible=visible, which=which, axis=axis, **kwargs)

    def _validate_plot_layout(self, layout_in, level=0):
        from geospacelab.datahub import VariableModel
        if (level == 0) and (not isinstance(layout_in, list)):
            raise TypeError("The plot layout must be a list!")
        type_layout_in = type(layout_in)
        layout_out = []
        for ind, elem in enumerate(layout_in):
            if isinstance(elem, list):
                layout_out.append(self._validate_plot_layout(elem, level=level+1))
            elif issubclass(elem.__class__, VariableModel):
                var = elem
                layout_out.append(var)
            elif isinstance(elem, int):
                index = elem
                var = self.variables[index]
                layout_out.append(var)
            else:
                raise TypeError
        return type_layout_in(layout_out)

    def save_figure(self, file_dir=None, file_name=None, append_time=True, dpi=300, file_format='png', **kwargs):
        if file_dir is None:
            file_dir = pathlib.Path().cwd()
        else:
            file_dir = pathlib.Path(file_dir)
        if type(file_name) is not str:
            raise ValueError

        if append_time:
            dt_range_str = self.get_dt_range_str(style='filename')
            file_name = '_'.join([file_name, dt_range_str])

        file_name = file_name + '.' + file_format

        plt.savefig(file_dir / file_name, dpi=dpi, format=file_format, **kwargs)

    def add_title(self, x=0.5, y=1.08, title=None, **kwargs):
        append_time = kwargs.pop('append_time', True)
        kwargs.setdefault('fontsize', plt.rcParams['figure.titlesize'])
        kwargs.setdefault('fontweight', 'roman')
        if append_time:
            dt_range_str = self.get_dt_range_str(style='title')
            title = title + ', ' + dt_range_str
        title = title.replace(', , ', ', ')
        super().add_title(x=x, y=y, title=title, **kwargs)

    def get_dt_range_str(self, style='title'):
        dt_fr = self._xlim[0]
        dt_to = self._xlim[1]
        if style == 'title':
            diff_days = dttool.get_diff_days(dt1=dt_fr, dt2=dt_to)
            if diff_days == 0:
                fmt1 = "%Y-%m-%dT%H:%M:%S"
                fmt2 = "%H:%M:%S"
            else:
                fmt1 = "%Y-%m-%dT%H:%M:%S"
                fmt2 = fmt1
            dt_range_str = dt_fr.strftime(fmt1) + ' - ' + dt_to.strftime(fmt2)
        elif style == 'filename':
            diff_days = dttool.get_diff_days(dt1=dt_fr, dt2=dt_to)
            if diff_days == 0:
                fmt1 = "%Y%m%d-%H%M%S"
                fmt2 = "%H%M%S"
            else:
                fmt1 = "%Y%m%d-%H%M%S"
                fmt2 = fmt1
            dt_range_str = dt_fr.strftime(fmt1) + '-' + dt_to.strftime(fmt2)
        return dt_range_str

    def add_vertical_line(self, dt_in, panel_index=-1,
                          label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):
        if type(dt_in) is not datetime.datetime:
            return

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('linewidth', 2)
        kwargs.setdefault('linestyle', '--')
        kwargs.setdefault('color', 'k')
        kwargs.setdefault('alpha', 0.8)

        if panel_index == -1:
            if 'major' not in self.extra_axes.keys():
                ax = self.add_major_axes()
            else:
                ax = self.extra_axes['major']
        else:
            ax = self.panels[panel_index].axes['major']

        xlim = self.panels[0].axes['major'].get_xlim()
        ylim = ax.get_ylim()
        diff_xlim = xlim[1] - xlim[0]
        diff_ylim = ylim[1] - ylim[0]
        x = mdates.date2num(dt_in)
        x = (x - xlim[0]) / diff_xlim

        y0 = 0 - bottom_extend
        y1 = 1. + top_extend
        line = mpl.lines.Line2D([x, x], [y0, y1], transform=ax.transAxes, **kwargs)
        line.set_clip_on(False)
        ax.add_line(line)

        if type(label) is str:
            if label_position == 'top':
                y = y1
                text_config.setdefault('va', 'bottom')
            elif label_position == 'bottom':
                y = y0
                text_config.setdefault('va', 'top')
            else:
                x = label_position[0]
                y = label_position[1]
                text_config.setdefault('va', 'center')
            text_config.setdefault('ha', 'center')
            text_config.setdefault('fontsize', plt.rcParams['axes.labelsize'])
            text_config.setdefault('fontweight', 'medium')
            text_config.setdefault('clip_on', False)
            ax.text(x, y, label, transform=ax.transAxes, **text_config)

    def add_horizontal_line(self, ys, panel_index=-1, *,
                          label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('linewidth', 1)
        kwargs.setdefault('linestyle', '--')
        kwargs.setdefault('color', 'k')
        kwargs.setdefault('alpha', 0.8)

        if panel_index == -1:
            ax_list = [p.axes['major'] for p in self.panels]
        else:
            ax_list = [self.panels[panel_index].axes['major']]

        if type(ys) is not list:
            ys = [ys] * len(ax_list)
        for ind, ax in enumerate(ax_list):
            xx = ax.get_xlim()
            yy = [ys[ind], ys[ind]]
            ax.plot(xx, yy, **kwargs)


    def search_UTs(self, search_step=1/86400, **kwargs) -> list:
        from scipy.interpolate import interp1d
        from scipy.signal import argrelmin
        import matplotlib.dates as mpl_dates
        panel = list(self.panels.values())[-1]
        var_for_config = panel._var_for_config
        x_depend = var_for_config.get_depend(axis=0, retrieve_data=True)
        x0 = np.array(mpl_dates.date2num(x_depend['UT'])).flatten()
        x_fr = x0[0]
        x_to = x0[-1]
        x1 = np.arange(x_fr, x_to, search_step)
        x1_0 = x1
        ind_0 = np.array(range(x1.size))
        for ind, (label, value) in enumerate(kwargs.items()):
            if label in x_depend.keys():
                y0 = x_depend[label].flatten()
            elif label in var_for_config.dataset.keys():
                y0 = var_for_config.dataset[label].value
            else:
                raise KeyError
            if ind == 0:
                if 'MLT' in label.upper():
                    y0_sin = np.sin(y0 / 24. * 2 * np.pi)
                    y0_cos = np.cos(y0 / 24. * 2 * np.pi)
                    itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                    itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                    y0_sin_i = itpf_sin(x1)
                    y0_cos_i = itpf_cos(x1)
                    rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                    rad = np.where((rad >= 0), rad, rad + 2 * np.pi)
                    y1 = rad / 2. / np.pi * 24.
                elif 'LON' in label.upper():
                    y0_sin = np.sin(y0 * np.pi / 180.)
                    y0_cos = np.cos(y0 * np.pi / 180.)
                    itpf_sin = interp1d(x0, y0_sin, bounds_error=False, fill_value='extrapolate')
                    itpf_cos = interp1d(x0, y0_cos, bounds_error=False, fill_value="extrapolate")
                    y0_sin_i = itpf_sin(x1)
                    y0_cos_i = itpf_cos(x1)
                    rad = np.sign(y0_sin_i) * (np.pi / 2 - np.arcsin(y0_cos_i))
                    y1 = rad * 180. / np.pi
                    if value is list:
                        value = [(v + 180.) % 360 - 180 for v in value]
                    else:
                        value = (value + 180.) % 360 - 180
                else:
                    itpf = interp1d(x0, y0, bounds_error=False, fill_value='extrapolate')
                    y1 = itpf(x1)
            else:
                y1 = y0
            if type(value) in (int, float):
                ind_1 = argrelmin(np.abs(y1 - value))
            elif type(value) is list:
                ind_1 = np.where((y1 > value[0]) & (y1 < value[1]))[0]
            else:
                raise NotImplementedError

            x1 = x1[ind_1]
            y1 = y1[ind_1]

        results = []
        for xx in x1:
            ind_1 = argrelmin(np.abs(x0 - xx))
            results.extend(mpl_dates.num2date(x0[ind_1]))

        return results

    def add_shading(self, dt_fr, dt_to, panel_index=0,
                    label=None, label_position=None, top_extend=0., bottom_extend=0., **kwargs):
        if type(dt_fr) is not datetime.datetime:
            return

        if label_position is None:
            label_position = 'top'
        text_config = kwargs.pop('text_config', {})
        kwargs.setdefault('edgecolor', 'none')
        kwargs.setdefault('facecolor', 'yellow')
        kwargs.setdefault('alpha', 0.4)

        if panel_index == 0:
            if 'major' not in self.extra_axes.keys():
                ax = self.add_major_axes()
            else:
                ax = self.extra_axes['major']
        else:
            ax = self.panels[panel_index].axes['major']

        xlim = self.panels[1].axes['major'].get_xlim()
        ylim = ax.get_ylim()
        diff_xlim = xlim[1] - xlim[0]
        diff_ylim = ylim[1] - ylim[0]
        x0 = mdates.date2num(dt_fr)
        x0 = (x0 - xlim[0]) / diff_xlim
        x1 = mdates.date2num(dt_to)
        x1 = (x1 - xlim[0]) / diff_xlim

        y0 = 0 - bottom_extend
        y1 = 1. + top_extend

        rectangle = mpl.patches.Rectangle((x0, y0), x1-x0, y1-y0, transform=ax.transAxes, **kwargs)
        rectangle.set_clip_on(False)
        ax.add_patch(rectangle)

        if type(label) is str:
            x = (x0 + x1) / 2
            if label_position == 'top':
                y = y1
                text_config.setdefault('va', 'bottom')
            elif label_position == 'bottom':
                y = y0
                text_config.setdefault('va', 'top')
            else:
                x = label_position[0]
                y = label_position[1]
                text_config.setdefault('va', 'center')
            text_config.setdefault('ha', 'center')
            text_config.setdefault('fontsize', plt.rcParams['axes.labelsize'])
            text_config.setdefault('fontweight', 'medium')
            text_config.setdefault('clip_on', False)
            ax.text(x, y, label, transform=ax.transAxes, **text_config)

    def add_top_bar(self, dt_fr, dt_to, bottom=0, top=0.02, facecolor='orange', **kwargs):
        bottom_extend = -1. - bottom
        top_extend = top
        kwargs.setdefault('alpha', 1.)
        self.add_shading(dt_fr, dt_to, bottom_extend=bottom_extend, top_extend=top_extend, facecolor=facecolor, **kwargs)




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

from geospacelab import preferences

from geospacelab.datahub import DatabaseModel


class SuperDARNDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


superdarn_database = SuperDARNDatabase('SuperDARN')
superdarn_database.url = 'http://vt.superdarn.org/tiki-index.php'
superdarn_database.category = 'online database'
superdarn_database.Notes = ''



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import cftime
import requests
import bs4
import pathlib
import re
import netCDF4 as nc
import pandas as pd
import ftplib
from contextlib import closing

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.datahub.sources.gfz.downloader as downloader
import geospacelab.datahub.sources.wdc as wdc
from geospacelab import preferences as prf
import geospacelab.datahub.sources.supermag.supermag_api as smapi

basekeys = ["sme", "sml", "smu", "mlat", "mlt", "glat", "glon", "stid", "num"]
# sunkeys: alias allowed of SUN___ -> ___s
sunkeys = ["smes", "smls", "smus", "mlats", "mlts", "glats", "glons", "stids", "nums"]
# darkkeys: alias allowed of DARK___ -> ___d
darkkeys = ["smed", "smld", "smud", "mlatd", "mltd", "glatd", "glond", "stidd", "numd"]
# regkeys: alias allowed of REGIONAL___ -> ___r
regkeys = ["smer", "smlr", "smur", "mlatr", "mltr", "glatr", "glonr", "stidr", "numr"]
pluskeys = ["smr", "ltsmr", "ltnum", "nsmr"]
indiceskeys = basekeys + sunkeys + darkkeys + regkeys + pluskeys
# 'all' means all the above

imfkeys = ["bgse", "bgsm", "vgse", "vgsm"]  # or imfall for all these
swikeys = ["pdyn", "epsilon", "newell", "clockgse", "clockgsm",
           "density"]  # % or swiall for all these

var_name_dict = {
    'sme': 'SME',
    "sml": 'SML',
    "smu": 'SMU',
    "mlat": 'AACGM_LAT',
    "mlt": 'AACGM_MLT',
    "glat": 'GEO_LAT',
    "glon": 'GEO_LON',
    "stid": 'STATION_ID',
    "num": 'STATION_NUM',
    'smes': 'SME_S',
    "smls": 'SML_S',
    "smus": 'SMU_S',
    "mlats": 'AACGM_LAT_S',
    "mlts": 'AACGM_MLT_S',
    "glats": 'GEO_LAT_S',
    "glons": 'GEO_LON_S',
    "stids": 'STATION_ID_S',
    "nums": 'STATION_NUM_S',
    'smed': 'SME_D',
    "smld": 'SML_D',
    "smud": 'SMU_D',
    "mlatd": 'AACGM_LAT_D',
    "mltd": 'AACGM_MLT_D',
    "glatd": 'GEO_LAT_D',
    "glond": 'GEO_LON_D',
    "stidd": 'STATION_ID_D',
    "numd": 'STATION_NUM_D',
    'smer': 'SME_R',
    "smlr": 'SML_R',
    "smur": 'SMU_R',
    "mlatr": 'AACGM_LAT_R',
    "mltr": 'AACGM_MLT_R',
    "glatr": 'GEO_LAT_R',
    "glonr": 'GEO_LON_R',
    "stidr": 'STATION_ID_R',
    "numr": 'STATION_NUM_R',
    "smr": 'SMR',
    "ltsmr": 'SMR_LT',
    "ltnum": 'LT_NUM',
    "nsmr": 'SMR_STATION_NUM',
}

class Downloader(object):

    def __init__(
            self,
            dt_fr: datetime.datetime,
            dt_to: datetime.datetime,
            user_id: str = None,
            products: str = None, # 'indicesall', 'swiall', 'imfall'
            force_download=True,
            **kwargs
    ):
        if products is None:
            products = ['indicesall']
        self.data_file_root_dir = prf.datahub_data_root_dir / 'SuperMAG' / 'INDICES'
        self.dt_fr = dt_fr
        self.dt_to = dt_to
        if user_id is None:
            self.user_id = prf.user_config['datahub']['supermag']['username']
        self.products = products

        self.force_download = force_download
        self.download()
        self.done = False

    def download(self):
        num_days = dttool.get_diff_days(self.dt_fr, self.dt_to)
        for nd in range(num_days+1):
            this_day = dttool.get_start_of_the_day(self.dt_fr) + datetime.timedelta(days=nd)
            extent = 86400.

            if self.products in [['all'], ['indicesall']]:
                product_name = 'indices'
            else:
                product_name = '_'.join(self.products)
            product_name = product_name.upper()
            file_name = '_'.join([
                'SuperMAG',
                product_name,
                this_day.strftime('%Y%m%d')
            ]) + '.nc'
            file_path = self.data_file_root_dir / this_day.strftime('%Y') / file_name
            if file_path.is_file() and not self.force_download:
                mylog.StreamLogger.info(f'The requested data file already exists! See "{file_path}".')
                self.done = True
                continue

            (status, idxdata) = smapi.SuperMAGGetIndices(
             self.user_id,
             this_day,
             extent,
             ','.join(self.products),
             FORMAT='list'
            )
            if status == 1:
                self.save_to_nc(idxdata, file_path)
            else:
                self.done = False
                mylog.StreamLogger.error(f'The requested data cannot be downloaded!')
                return
    def save_to_nc(self, idxdata, file_path):
        def sm_t_to_datetime(tval):
            jd = (tval / 86400.0) + 2440587.5
            timestamp = pd.to_datetime(jd, unit='D', origin='julian')  # format YYYY-MM-DD HH:MM:SS.ssssss
            return timestamp.to_pydatetime()
        vars = {}
        dts = []
        dt_0 = dttool.get_start_of_the_day(sm_t_to_datetime(idxdata[0]['tval']))
        for item in idxdata:
            dt_c = sm_t_to_datetime(item['tval'])
            dt_c = dt_0 + datetime.timedelta(minutes=np.around((dt_c - dt_0).total_seconds() / 60))
            dts.append(dt_c)

            for key in item.keys():
                if key == 'tval':
                    continue
                value = item[key]
                if type(value) is list:
                    nelem = len(value)
                else:
                    nelem = 1
                    value = [value]
                value = np.array(value)[np.newaxis, :]
                vars.setdefault(key, np.empty((0, nelem), dtype=type(value[0, 0])))
                vars[key] = np.vstack((vars[key], value))
        keys = list(vars.keys())
        for key in keys:
            if 'mlat' in key:
                new_key = key.replace('mlat', '_AACGM_LAT')
            elif 'mlt' in key:
                new_key = key.replace('mlt', '_AACGM_MLT')
            elif 'glat' in key:
                new_key = key.replace('glat', '_GEO_LAT')
            elif 'glon' in key:
                new_key = key.replace('glon', '_GEO_LON')
            elif 'stid' in key:
                new_key = key.replace('stid', '_STATION')
            elif 'smr' in key:
                new_key = key.replace('smr', 'SMR')
            else:
                new_key = key
            if 'num' in new_key:
                new_key = key.replace('num', '_NUM')
            vars[new_key] = vars.pop(key)

        file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
        fnc = nc.Dataset(file_path, 'w')
        fnc.title = "SuperMAG Indices (Daily)"

        num_rows = len(dts)
        time_array = np.array(cftime.date2num(dts, units='seconds since 1970-01-01 00:00:00.0'))
        fnc.createDimension('UNIX_TIME', num_rows)
        fnc.createDimension('SME_R_NUM', vars['SMEr_NUM'].shape[1])
        time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
        time.units = 'seconds since 1970-01-01 00:00:00.0'
        time[::] = time_array[::]

        for key in vars.keys():
            var = vars[key]
            if var.shape[1] == 1:
                dim = ('UNIX_TIME', )
            else:
                dim = ('UNIX_TIME', 'SME_R_NUM')
            var_nc = fnc.createVariable(key, var.dtype, dim)
            var_nc[::] = var[::]
        mylog.StreamLogger.info(
            "The requested SuperMAG data has been saved in the file {}.".format(file_path))
        fnc.close()
        self.done = True


if __name__ == "__main__":
    dt_fr1 = datetime.datetime(2012, 1, 1)
    dt_to1 = datetime.datetime(2012, 1, 16)
    Downloader(dt_fr1, dt_to1)






# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import numpy as np
import datetime as dt
from matplotlib.ticker import LogFormatterSciNotation
import geospacelab.toolbox.utilities.pydatetime as dttool
import os

from dateutil.rrule import (rrule, MO, TU, WE, TH, FR, SA, SU, YEARLY,
                            MONTHLY, WEEKLY, DAILY, HOURLY, MINUTELY,
                            SECONDLY)

HOURS_PER_DAY = 24.
MIN_PER_HOUR = 60.
SEC_PER_MIN = 60.
MONTHS_PER_YEAR = 12.

DAYS_PER_WEEK = 7.
DAYS_PER_MONTH = 30.
DAYS_PER_YEAR = 365.0

MINUTES_PER_DAY = MIN_PER_HOUR * HOURS_PER_DAY

SEC_PER_HOUR = SEC_PER_MIN * MIN_PER_HOUR
SEC_PER_DAY = SEC_PER_HOUR * HOURS_PER_DAY
SEC_PER_WEEK = SEC_PER_DAY * DAYS_PER_WEEK


class DatetimeMajorLocator(mdates.AutoDateLocator):
    def __init__(self, tz=None, minticks=5, maxticks=12, interval_multiples=True):

        super().__init__(tz=tz, minticks=minticks, maxticks=maxticks, interval_multiples=interval_multiples)
        self.intervald[MINUTELY] = [1, 3, 5, 10, 15, 30]
        self.intervald[SECONDLY] = [1, 5, 10, 15, 30]


class DatetimeMinorLocator(mdates.AutoDateLocator):

    def __init__(self, ax=None, majorlocator=None, tz=None, minticks=12, maxticks=50, interval_multiples=True):

        if majorlocator is not None:
            tz = majorlocator.tz
        # if ax is not None:
        #    minticks = 2 * len(ax.xaxis.get_major_ticks())
        #    maxticks = 8 * len(ax.xaxis.get_major_ticks())
        super().__init__(tz=tz, minticks=minticks, maxticks=maxticks, interval_multiples=interval_multiples)


class DatetimeMajorFormatter(mdates.AutoDateFormatter):
    def __init__(self, locator, scaled: dict or None = None, tz=None, defaultfmt='%Y-%m-%d', *, usetex=True):
        super().__init__(locator, tz=None, defaultfmt=defaultfmt, usetex=usetex)
        self.scaled[1 / HOURS_PER_DAY] = formatter_hour_per_day
        self.scaled[1 / MINUTES_PER_DAY] = formatter_minute_per_day
        if scaled is not None:
           self.scaled.update(**scaled)


def formatter_hour_per_day(x, pos):
    dtx = mpl.dates.num2date(x)
    dtx = dtx.replace(tzinfo=None)
    delta = dtx - dt.datetime(dtx.year, dtx.month, dtx.day)
    if delta.total_seconds() == 0:
        fmt1 = "%b %d"
    else:
        fmt1 = "%H:%M"
    return dtx.strftime(fmt1)


def formatter_minute_per_day(x, pos):
    dtx = mpl.dates.num2date(x)
    dtx = dtx.replace(tzinfo=None)
    delta = dtx - dt.datetime(dtx.year, dtx.month, dtx.day)
    if delta.total_seconds() == 0:
        fmt1 = "%b %d"
    else:
        fmt1 = "%H:%M"
    return dtx.strftime(fmt1)



def set_timeline(dt_start, dt_stop, **kwargs):
    mdlocators = {
        1:   'mdates.MicrosecondLocator',
        2:   'mdates.SecondLocator',
        3:   'mdates.MinuteLocator',
        4:   'mdates.HourLocator',
        5:   'mdates.DayLocator',
        6:   'mdates.MonthLocator',
        7:   'mdates.YearLocator',
    }
    default_second = {
        'range':        [1, 2, 3, 5, 10, 15, 20, 30, 40, 60],
        'majorscale':   [10/60, 15/60, 30/60, 1, 2, 3, 4, 5, 5, 10],
        'minorscale':   [2/60, 2/60, 5/60, 10/60, 20/60, 30/60, 1, 1, 1, 2],
        'scale':        1000
    }
    default_minute = {
        'range':        [1, 2, 3, 5, 8, 15, 20, 30, 40, 60],
        'majorscale':   [15/60, 20/60, 30/60, 1, 2, 3, 4, 5, 10],
        'minorscale':   [3/60, 5/60, 5/60, 10/60, 20/60, 30/60, 1, 1, 2],
        'scale': 60
    }
    default_hour = {
        'range':        [1, 2, 3, 5, 8, 12, 18, 24],
        'majorscale':   [15/60, 20/60, 30/60, 1, 2, 3, 4],
        'minorscale':   [3/60, 5/60, 5/60, 10/60, 20/60, 30/60, 1],
        'scale': 60
    }

    default_day = {
        'range':      [1, 2, 3, 5, 8, 14, 21, 30, 50, 80, 122],
        'majorscale': [6/24, 8/24, 12/24, 1, 2, 3, 5, 10, 15],
        'minorscale': [1/24, 2/24, 3/24, 4/24, 8/24, 12/24, 1, 2, 3],
        'scale': 24
    }

    default_month = {
        'range':         [3, 4, 7, 12],
        'majorscale':    [1/2, 1, 2],
        'minorscale':    [1/6, 1/3, 1/2],
        'scale':         30
    }

    default_year = {
        'range':            [1, 2, 3, 5, 10, 15, 20, 30, 50, 100, 200, 400, 800],
        'majorscale':      [3/12, 4/12, 6/12, 1, 2, 3, 5, 5, 10, 20, 50, 100],
        'minorscale':      [1/12, 1/12, 2/12, 3/12, 4/12, 6/12, 1, 1, 2, 5, 10, 20],
        'scale': 12
    }

    if 'my_second' in kwargs.keys():
        default_second.update(kwargs['my_second'])
    if 'my_minute' in kwargs.keys():
        default_second.update(kwargs['my_minute'])
    if 'my_hour' in kwargs.keys():
        default_second.update(kwargs['my_hour'])
    if 'my_day' in kwargs.keys():
        default_second.update(kwargs['my_day'])
    if 'my_month' in kwargs.keys():
        default_second.update(kwargs['my_month'])
    if 'my_year' in kwargs.keys():
        default_second.update(kwargs['my_year'])

    default_settings = {
        1:      {},
        2:      default_second,
        3:      default_minute,
        4:      default_hour,
        5:      default_day,
        6:      default_month,
        7:      default_year
    }

    tdelta = dt_stop - dt_start
    diff = tdelta.total_seconds()
    for ind in range(2, 8):
        range_ = default_settings[ind]['range']
        if (diff >= range_[0]) and (diff < range_[-1]):
            break
        else:
            if ind == 2:
                diff = diff/60
            elif ind ==3:
                diff = diff/60
            elif ind == 4:
                diff = diff/24
            elif ind == 5:
                diff = dttool.get_diff_months(dt_start, dt_stop)
            elif ind == 6:
                diff = dt_stop.year - dt_start.year

    setting = default_settings[ind]
    range_ = setting['range']
    for range_ind, value in enumerate(range_):
        if diff < value:
            break
    range_ind = range_ind - 1
    majorscale = setting['majorscale'][range_ind]
    minorscale = setting['minorscale'][range_ind]
    if majorscale < range_[0]:
        majorlocatorclass = eval(mdlocators[ind - 1])
        majorscale = majorscale * setting['scale']
    else:
        majorlocatorclass = eval(mdlocators[ind])
    if minorscale < range_[0]:
        minorlocatorclass = eval(mdlocators[ind - 1])
        minorscale = minorscale * setting['scale']
    else:
        minorlocatorclass = eval(mdlocators[ind])
    majorscale = int(majorscale)
    minorscale = int(minorscale)
    # for microseconds,
    if majorlocatorclass is mdates.MicrosecondLocator:
        interval = majorscale
        majorlocator = majorlocatorclass(interval=majorscale)
        if dt_start.minute != dt_stop.minute:
            fmt = "%M:%S.%f"
        else:
            fmt = "%S.%f"

        def formatter_microsecond(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, dt.month, dtx.day, dtx.hour, dtx.minute, 0)
            if delta.total_seconds() == 0:
                fmt1 = "%M:%S.%f"
            else:
                fmt1 = "%S.%f"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_microsecond)
    if minorlocatorclass is mdates.MicrosecondLocator:
        interval = minorscale
        minorlocator = minorlocatorclass(interval=minorscale)

    # for second
    if majorlocatorclass is mdates.SecondLocator:
        by1 = range(0, 60, majorscale)
        majorlocator = majorlocatorclass(bysecond=by1, interval=1)
        if dt_start.hour != dt_stop.hour:
            fmt = "%H:%M:%S"
        else:
            fmt = "%M:%S"

        def formatter_second(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, dtx.month, dtx.day, dtx.hour, 0, 0)
            if delta.total_seconds() == 0:
                fmt1 = "%H:%M:%S"
            else:
                fmt1 = "%M:%S"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_second)
    if minorlocatorclass is mdates.SecondLocator:
        by1 = range(0, 60, minorscale)
        minorlocator = minorlocatorclass(bysecond=by1, interval=1)

    # for minute
    if majorlocatorclass is mdates.MinuteLocator:
        by1 = range(0, 60, majorscale)
        majorlocator = majorlocatorclass(byminute=by1, interval=1)
        if dt_start.day != dt_stop.day:
            fmt = "%d %H:%M"
        else:
            fmt = "%H:%M"

        def formatter_minute(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, dtx.month, dtx.day, 0, 0, 0)
            if delta.total_seconds() == 0:
                fmt1 = "%b %d"
            else:
                fmt1 = "%H:%M"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_minute)
    if minorlocatorclass is mdates.MinuteLocator:
        by1 = range(0, 60, minorscale)
        minorlocator = minorlocatorclass(byminute=by1, interval=1)

    # for hour
    if majorlocatorclass is mdates.HourLocator:
        by1 = range(0, 24, majorscale)
        majorlocator = majorlocatorclass(byhour=by1, interval=1)
        if dt_start.month != dt_stop.month:
            fmt = "%d/%m %H"
        elif dt_start.day != dt_stop.day:
            fmt = "%d %H:%M"
        else:
            fmt = "%H:%M"

        def formatter_hour(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, dtx.month, dtx.day)
            if delta.total_seconds() == 0:
                fmt1 = "%b %d"
            else:
                fmt1 = "%H:%M"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_hour)
    if minorlocatorclass is mdates.HourLocator:
        by1 = range(0, 24, minorscale)
        minorlocator = minorlocatorclass(byhour=by1, interval=1)

    # for day
    if majorlocatorclass is mdates.DayLocator:
        temp = int(np.floor(31.5/majorscale))
        by1 = range(1, temp*majorscale, majorscale)
        majorlocator = majorlocatorclass(bymonthday=by1, interval=1)
        if dt_start.year != dt_stop.year:
            fmt = "%Y-%m-%d"
        else:
            fmt = "%b %d"

        def formatter_day(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, 1, 1)
            if delta.total_seconds() == 0:
                fmt1 = "%Y-%m-%d"
            else:
                fmt1 = "%m-%d"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_day)
    if minorlocatorclass is mdates.DayLocator:
        temp = int(np.floor(31.5 / minorscale))
        by1 = range(1, temp * minorscale, minorscale)
        minorlocator = minorlocatorclass(bymonthday=by1, interval=1)

    # for month
    if majorlocatorclass is mdates.MonthLocator:
        by1 = range(1, 13, majorscale)
        majorlocator = majorlocatorclass(bymonth=by1, interval=1)
        fmt = "%Y-%m"

        def formatter_month(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            delta = dtx - dt.datetime(dtx.year, 1, 1)
            if delta.total_seconds() == 0:
                fmt1 = "%Y-%m-%d"
            else:
                fmt1 = "%m-%d"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_month)
    if minorlocatorclass is mdates.MonthLocator:
        by1 = range(1, 13, minorscale)
        minorlocator = minorlocatorclass(bymonth=by1, interval=1)

    # for year
    if majorlocatorclass is mdates.YearLocator:
        majorlocator = majorlocatorclass(base=majorscale)

        def formatter_year(x, pos):
            dtx = mpl.dates.num2date(x)
            dtx = dtx.replace(tzinfo=None)
            fmt1 = "%Y"
            return dtx.strftime(fmt1)
        func_formatter = mpl.ticker.FuncFormatter(formatter_year)
    if minorlocatorclass is mdates.YearLocator:
        minorlocator = minorlocatorclass(base=minorscale)
    majorformatter = mdates.DateFormatter(fmt)
    majorformatter = func_formatter
    return majorlocator, minorlocator, majorformatter


# def timeline_format_function(x, pos=None):
#     x = mpl.dates.num2date(x)
#     if pos == 0:
#         # fmt = '%D %H:%M:%S.%f'
#         fmt = '%H:%M'
#     else:
#       fmt = '%H:%M'
#     label = x.strftime(fmt)
#     #label = label.rstrip("0")
#     #label = label.rstrip(".")
#     return label

#
# def set_ticks_datetime(fig, ax, axis='x', locator=None, formatter=None, interval=None, visable='on'):
#     xlim = getattr(ax, 'get_' + axis + 'lim')()
#     dxlim = np.diff(xlim) * 86400.
#     dsec = dxlim
#     dmin = dsec / 60.
#     dhour = dsec / 60.
#     dday = dhour / 24.
#     dmonth = dday / 30.
#     dyear = dday / 365.
#     if locator is None:
#         pass

#    locObj = getattr(mdates, locator)
    # majorlocator = mdates.MinuteLocator(interval=1)
    # formatter = mdates.AutoDateFormatter(dtlocator)
    # formatter.scaled[1 / (24. * 60.)] = matplotlib.ticker.FuncFormatter(my_format_function)
    # if iax < nax - 1:
    #     formatter = matplotlib.ticker.NullFormatter()
    # ax.xaxis.set_major_locator(dtlocator)
    # ax.xaxis.set_major_formatter(formatter)
    # ax.xaxis.set_minor_locator(mdates.SecondLocator(interval=10))
    # ax.xaxis.set_tick_params(labelsize='small')



# class CustomTicker(LogFormatterSciNotation):
#     # https://stackoverflow.com/questions/43923966/logformatter-tickmarks-scientific-format-limits
#     def __call__(self, x, pos=None):
#         if x not in [0.1,1,10]:
#             return LogFormatterSciNotation.__call__(self,x, pos=None)
#         else:
#             return "{x:g}".format(x=x)
# fig = plt.figure(figsize=[7,7])
# ax = fig.add_subplot(111)
# ax.set_yscale('log')
# ax.set_xscale('log')
# ax.plot(np.logspace(-4,4), np.logspace(-4,4))
#
# ax.xaxis.set_major_formatter(CustomTicker())
# plt.show()


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import toml
import pathlib
import time
import os
import geospacelab.toolbox.utilities.pylogging as mylog

# datahub_data_root_dir = pathlib.Path("/Users/lcai/01-Work/00-Data/")
# datahub_data_root_dir = pathlib.Path("/home/lei/afys-data")

package_name = "geospacelab"

on_rtd = os.environ.get('READTHEDOCS') == 'True'


class Preferences(object):
    _on_rtd = on_rtd

    def __init__(self):
        self.package_name = package_name

        self.user_config = {}
        self.set_user_config()

        if self._on_rtd:
            self.datahub_data_root_dir = pathlib.Path.home() / 'Geospacelab' / 'Data'
        else:
            try:
                self.datahub_data_root_dir = self.user_config['datahub']['data_root_dir']
            except KeyError:
                self.datahub_data_root_dir = None

    @property
    def datahub_data_root_dir(self):
        return self._datahub_data_root_dir

    @datahub_data_root_dir.setter
    def datahub_data_root_dir(self, path):
        if path == '':
            path = None
        if path is None:
            data_path = pathlib.Path.home() / 'Geospacelab' / 'Data'
            time.sleep(0.2)
            result = input(f"Store the data in the default directory ({data_path})? [y]/n: ")
            time.sleep(0.2)
            if result.lower() in ['', 'y', 'yes']:
                self._datahub_data_root_dir = data_path
            elif result.lower() in ['n', 'no']:
                time.sleep(0.2)
                result = input('Input the root directory for storing data: ')
                time.sleep(0.2)
                if str(result):
                    self._datahub_data_root_dir = pathlib.Path(result)
                else:
                    raise NotADirectoryError('"Set the default root directory in ~/.geospacelab/config.toml!"')
            else:
                raise ValueError("The input is not correct!")
        else:
            self._datahub_data_root_dir = pathlib.Path(path)

        if not self._datahub_data_root_dir.is_dir():
            mylog.simpleinfo.info(f"The direction ({self._datahub_data_root_dir}) does not exist!")
            if self._on_rtd:
                result = 'y'
            else:
                time.sleep(0.2)
                result = input("Create it? [y]/n: ")
                time.sleep(0.2)
            if result.lower() in ['', 'y', 'yes']:
                self._datahub_data_root_dir.mkdir(parents=True, exist_ok=False)
            elif result.lower() in ['n', 'no']:
                mylog.simpleinfo.info("The directory was not created! Set the default root directory in ~/.geospacelab/config.toml!")
            else:
                raise ValueError("The input is not correct!")
        else:
            if path is None:
                mylog.simpleinfo.info(f"Note: The directory ({self._datahub_data_root_dir}) has existed!")

        self.user_config['datahub']['data_root_dir'] = str(self._datahub_data_root_dir)
        if path is None:
            self.set_user_config(user_config=self.user_config, set_as_default=True)

    def set_user_config(self, user_config=None, set_as_default=False):

        home_dir = pathlib.Path.home()

        config_file_path = home_dir / ('.' + package_name) / 'config.toml'

        config_file_path.parent.mkdir(exist_ok=True, parents=True)

        if not config_file_path.is_file():
            config_string = """
# This is a document for the user's configuration in the TOML format.
package_name = \"{}\"
[datahub]
data_root_dir = \"\"
            """.format(package_name)
            mylog.StreamLogger.info(
                "A toml file for user's configuration has been created at {}.".format(config_file_path)
            )
            time.sleep(2)
            config_file_path.touch()
            with open(config_file_path, 'w') as f:
                f.write(config_string)

        user_config_dict = {}
        if user_config is not None:
            if isinstance(user_config, str):
                user_config_dict = toml.loads(user_config)
            elif isinstance(user_config, dict):
                user_config_dict = user_config
            if set_as_default:
                with open(config_file_path, 'w') as f:
                    toml.dump(user_config, f)
        else:
            user_config_dict = toml.load(config_file_path)
        self.user_config = user_config_dict
        return user_config_dict



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import requests
import bs4
import re

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab import preferences as prf


class Downloader(object):

    def __init__(self, dt_fr, dt_to, direct_download=True, force_download=True):

        self.url_base = "https://cdaweb.gsfc.nasa.gov/pub/data/"
        self.force_download = force_download

        self.dt_fr = dt_fr
        self.dt_to = dt_to
        self.source_file_paths = []

        if direct_download:
            self.download()

    def download(self):
        raise NotImplementedError

    def search_file(self, subdirs, file_name_patterns, logging=True):
        url = self.url_base + '/'.join(subdirs)

        r = requests.get(url)

        soup = bs4.BeautifulSoup(r.text, 'html.parser')
        a_tags = soup.find_all('a', href=True)
        hrefs = [a_tag['href'] for a_tag in a_tags]

        search_pattern = '.*' + '.*'.join(file_name_patterns) + '.*'
        fn_regex = re.compile(search_pattern)
        hrefs = list(filter(fn_regex.match, hrefs))

        if logging:
            if len(hrefs) == 0:
                mylog.StreamLogger.warning("Cannot find any files matching the file patterns!")
            elif len(hrefs) > 1:
                mylog.StreamLogger.warning("Multiple files matching the file patterns are found!")

        for href in hrefs:
            self.source_file_paths.append(url + href)
        return

    def save_file(self, file_dir, file_name, r_source_file):
        file_path = file_dir / file_name

        if file_path.is_file():
            mylog.simpleinfo.info(
                "The file {} exists in the directory {}.".format(file_path.name, file_path.parent.resolve()))

        if self.force_download:
            file_path.parent.resolve().mkdir(parents=True, exist_ok=True)
            with open(file_path, "wb") as f:
                mylog.simpleinfo.info(
                    "Downloading {} to the directory {} ...".format(file_path.name, file_path.parent.resolve())
                )
                f.write(r_source_file.content)
                mylog.simpleinfo.info("Done")




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

from geospacelab import preferences

from geospacelab.datahub import DatabaseModel


class ISEEDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


isee_database = ISEEDatabase('ISEE')
isee_database.url = 'https://stdb2.isee.nagoya-u.ac.jp/index_e.html'
isee_database.category = 'online database'
isee_database.Notes = """
The global TEC data were derived from GNSS observation data in Receiver INdependent EXchange (RINEX) format obtained from 
a lot of regional GNSS reciever network all over the world. The number of GNSS stations reached more than 8800 in April 
2019. These GNSS observation data were provided by many data providers (data provider list). If you use GNSS-TEC data 
for the purpose of scientific research, you should include the following acknowledgment text.

Global GNSS-TEC data processing has been supported by JSPS KAKENHI Grant Number 16H06286. GNSS RINEX files for the 
GNSS-TEC processing are provided from many organizations listed by the webpage
(http://stdb2.isee.nagoya-u.ac.jp/GPS/GPS-TEC/gnss_provider_list.html).
"""


# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


"""
Utilities related to the system's directories and files
"""
import os
from pathlib import Path


def path_join(*args):
    """
    A wrapper of os.path.join() returns the full directory as a str.

            Parameters:
                    args: list  # Any member (str) in args.

            Returns:
                    new_path: str # Joined directory
    """
    new_path = os.path.join(*args)
    return new_path





# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import pathlib
import h5py
import numpy as np

import geospacelab.toolbox.utilities.pylogging as mylog

# define the default variable name dictionary
default_variable_name_dict = {
    'JN_e': 'EL_I_FLUX',
    'JN_i': 'ION_I_FLUX',
    'JE_e': 'EL_I_ENER',
    'JE_i': 'ION_I_ENER',
    'E_e_MEAN': 'EL_M_ENER',
    'E_i_MEAN': 'ION_M_ENER',
    'jE_e': 'EL_D_ENER',
    'jE_i': 'ION_D_ENER',
    'jN_e': 'EL_D_FLUX',
    'jN_i': 'ION_D_FLUX',
    'SC_GEO_LAT': 'GDLAT',
    'SC_GEO_LON': 'GLON',
    'SC_GEO_ALT': 'GDALT',
    'SC_MAG_LAT': 'MLAT',
    'SC_MAG_LON': 'MLONG',
    'SC_MAG_MLT': 'MLT',
    'CH_CTRL_E':  'CH_CTRL_ENER',
}


class Loader(object):
    """
    :param file_path: the full path of the data file
    :type file_path: pathlib.Path or str
    :param file_type: the type of the file, [cdf]
    :param variable_name_dict: the dictionary for mapping the variable names from the cdf files to the dataset
    :type variable_name_dict: dict
    :param direct_load: call the method :meth:`~.LoadModel.load_data` directly or not
    :type direct_load: bool
    """
    def __init__(self, file_path, file_type='hdf5', variable_name_dict=None, direct_load=True, **kwargs):

        self.file_path = pathlib.Path(file_path)
        self.file_ext = file_type
        self.variables = {}
        self.metadata = {}

        if variable_name_dict is None:
            variable_name_dict = default_variable_name_dict
        self.variable_name_dict = variable_name_dict

        if direct_load:
            self.load_data()

    def load_data(self, **kwargs):
        if self.file_ext == 'hdf5':
            self.load_hdf5_data()

    def load_hdf5_data(self):
        """
        load the data from the cdf file
        :return:
        """
        with h5py.File(self.file_path, 'r') as fh5:
            # load metadata
            metadata = {}

            # load data
            data_parameters = list(zip(*tuple(fh5['Data']['Array Layout']['1D Parameters']['Data Parameters'][:])))
            data = fh5['Data']['Array Layout']['1D Parameters']
            var_names_h5 = [vn.decode('UTF-8') for vn in data_parameters[0]]
            nvar_h5 = len(var_names_h5)
            vars_h5 = {}
            nrow = data['gdlat'][:].shape[0]
            for ip in range(nvar_h5):
                vars_h5[var_names_h5[ip]] = np.array(data[var_names_h5[ip].lower()]).reshape((nrow, 1))

            data_parameters = list(zip(*tuple(fh5['Data']['Array Layout']['2D Parameters']['Data Parameters'][:])))
            data = fh5['Data']['Array Layout']['2D Parameters']
            var_names_h5 = [vn.decode('UTF-8') for vn in data_parameters[0]]
            nvar_h5 = len(var_names_h5)
            for ip in range(nvar_h5):
                vars_h5[var_names_h5[ip]] = np.array(data[var_names_h5[ip].lower()]).T

            for var_name, var_name_h5 in self.variable_name_dict.items():
                try:
                    self.variables[var_name] = vars_h5[var_name_h5]
                except KeyError:
                    print(f"{var_name} is None!")
                    self.variables[var_name] = None

            # channel grid
            energy_spacing = vars_h5['CH_CTRL_ENER'] / 2
            E_chs = fh5['Data']['Array Layout']['ch_energy'][:]
            E_chs = np.array(E_chs).reshape(1, E_chs.shape[0])
            newchannel = E_chs - energy_spacing
            boudarychannel = E_chs[0, -1] + energy_spacing[:, -1]
            boudarychannel = boudarychannel.reshape((boudarychannel.shape[0], 1))
            self.variables['ENERGY_CHANNEL_GRID'] = np.hstack((newchannel, boudarychannel))

            # add datetime
            timestamps = fh5['Data']['Array Layout']['timestamps'][:]
            dt0 = datetime.datetime(1970, 1, 1, 0, 0, 0)

            dts = np.array([dt0 + datetime.timedelta(seconds=secs)
                               for ind, secs in enumerate(timestamps)]).reshape(nrow, 1)
            self.variables['SC_DATETIME'] = dts

            self.variables['JN_e'][self.variables['JN_e'] < 1.] = np.nan
            self.variables['JN_i'][self.variables['JN_i'] < 1.] = np.nan
            self.variables['JE_e'][self.variables['JE_e'] < 1.] = np.nan
            self.variables['JE_i'][self.variables['JE_i'] < 1.] = np.nan
            self.variables['jN_e'][self.variables['jN_e'] < 1.] = np.nan
            self.variables['jN_i'][self.variables['jN_i'] < 1.] = np.nan
            self.variables['jE_e'][self.variables['jE_e'] < 1.] = np.nan
            self.variables['jE_i'][self.variables['jE_i'] < 1.] = np.nan